@@ -72,11 +72,13 @@ static struct repo_write {
 	struct got_repository *repo;
 	int *pack_fds;
 	int *temp_fds;
+	FILE *base_file;
+	FILE *accum_file;
 	int session_fd;
 	struct gotd_imsgev session_iev;
-	struct got_pathlist_head *protected_tag_namespaces;
-	struct got_pathlist_head *protected_branch_namespaces;
-	struct got_pathlist_head *protected_branches;
+	struct got_pathlist_head protected_tag_namespaces;
+	struct got_pathlist_head protected_branch_namespaces;
+	struct got_pathlist_head protected_branches;
 	struct {
 		FILE *f1;
 		FILE *f2;
@@ -84,6 +86,10 @@ static struct repo_write {
 		int fd2;
 	} diff;
 	int refs_listed;
+	int have_packfile;
+	struct got_pathlist_head *protected_refs_cur;
+	size_t nprotected_refs_needed;
+	size_t nprotected_refs_received;
 } repo_write;
 
 struct gotd_ref_update {
@@ -101,6 +107,7 @@ static struct repo_write_client {
 	int				 fd;
 	int				 pack_pipe;
 	struct got_pack			 pack;
+	struct got_packidx		 *packidx;
 	uint8_t				 pack_sha1[SHA1_DIGEST_LENGTH];
 	int				 packidx_fd;
 	struct gotd_ref_updates		 ref_updates;
@@ -163,7 +170,7 @@ send_peeled_tag_ref(struct got_reference *ref, struct got_object *obj,
 		goto done;
 	}
 
-	wbuf = imsg_create(ibuf, GOTD_IMSG_REF, PROC_REPO_WRITE,
+	wbuf = imsg_create(ibuf, GOTD_IMSG_REF, GOTD_PROC_REPO_WRITE,
 	    repo_write.pid, len);
 	if (wbuf == NULL) {
 		err = got_error_from_errno("imsg_create REF");
@@ -211,7 +218,7 @@ send_ref(struct got_reference *ref, struct imsgbuf *ibuf)
 	if (err)
 		return err;
 
-	wbuf = imsg_create(ibuf, GOTD_IMSG_REF, PROC_REPO_WRITE,
+	wbuf = imsg_create(ibuf, GOTD_IMSG_REF, GOTD_PROC_REPO_WRITE,
 	    repo_write.pid, len);
 	if (wbuf == NULL) {
 		err = got_error_from_errno("imsg_create REF");
@@ -303,7 +310,7 @@ list_refs(struct imsg *imsg)
 			irefs.nrefs++;
 	}
 
-	if (imsg_compose(&ibuf, GOTD_IMSG_REFLIST, PROC_REPO_WRITE,
+	if (imsg_compose(&ibuf, GOTD_IMSG_REFLIST, GOTD_PROC_REPO_WRITE,
 	    repo_write.pid, -1, &irefs, sizeof(irefs)) == -1) {
 		err = got_error_from_errno("imsg_compose REFLIST");
 		goto done;
@@ -462,7 +469,8 @@ protect_require_yca(struct got_object_id *tip_id,
 			break;
 		}
 
-		if (got_object_idset_num_elements(traversed_set) >=
+		if (max_commits_to_traverse > 0 &&
+		    got_object_idset_num_elements(traversed_set) >=
 		    max_commits_to_traverse)
 			break;
 
@@ -704,7 +712,7 @@ recv_ref_update(struct imsg *imsg)
 		}
 	}
 
-	gotd_imsg_send_ack(&ref_update->new_id, &ibuf, PROC_REPO_WRITE,
+	gotd_imsg_send_ack(&ref_update->new_id, &ibuf, GOTD_PROC_REPO_WRITE,
 	    repo_write.pid);
 
 	ref_update->ref = ref;
@@ -834,12 +842,13 @@ copy_ref_delta(int infd, int outfd, off_t *outsize, BUF *buf, size_t *buf_pos,
 		return err;
 
 	*buf_pos += SHA1_DIGEST_LENGTH;
+	*outsize += SHA1_DIGEST_LENGTH;
 	return NULL;
 }
 
 static const struct got_error *
-copy_offset_delta(int infd, int outfd, off_t *outsize, BUF *buf, size_t *buf_pos,
-    struct got_hash *ctx)
+copy_offset_delta(int infd, int outfd, off_t *outsize, BUF *buf,
+    size_t *buf_pos, struct got_hash *ctx)
 {
 	const struct got_error *err = NULL;
 	uint64_t o = 0;
@@ -1177,8 +1186,8 @@ report_pack_status(const struct got_error *unpack_err)
 		istatus.reason_len = strlen(unpack_ok);
 
 	len = sizeof(istatus) + istatus.reason_len;
-	wbuf = imsg_create(&ibuf, GOTD_IMSG_PACKFILE_STATUS, PROC_REPO_WRITE,
-	    repo_write.pid, len);
+	wbuf = imsg_create(&ibuf, GOTD_IMSG_PACKFILE_STATUS,
+	    GOTD_PROC_REPO_WRITE, repo_write.pid, len);
 	if (wbuf == NULL) {
 		err = got_error_from_errno("imsg_create PACKFILE_STATUS");
 		goto done;
@@ -1362,10 +1371,9 @@ done:
 static const struct got_error *
 verify_packfile(void)
 {
-	const struct got_error *err = NULL, *close_err;
+	const struct got_error *err = NULL;
 	struct repo_write_client *client = &repo_write_client;
 	struct gotd_ref_update *ref_update;
-	struct got_packidx *packidx = NULL;
 	struct stat sb;
 	char *id_str = NULL;
 	struct got_object *obj = NULL;
@@ -1389,13 +1397,18 @@ verify_packfile(void)
 	if (fstat(client->packidx_fd, &sb) == -1)
 		return got_error_from_errno("pack index fstat");
 
-	packidx = malloc(sizeof(*packidx));
-	memset(packidx, 0, sizeof(*packidx));
-	packidx->fd = client->packidx_fd;
-	client->packidx_fd = -1;
-	packidx->len = sb.st_size;
+	if (client->packidx != NULL)
+		return got_error(GOT_ERR_PRIVSEP_MSG);
 
-	err = got_packidx_init_hdr(packidx, 1, client->pack.filesize);
+	client->packidx = calloc(1, sizeof(*client->packidx));
+	if (client->packidx == NULL)
+		return got_error_from_errno("calloc");
+
+	client->packidx->fd = client->packidx_fd;
+	client->packidx_fd = -1;
+	client->packidx->len = sb.st_size;
+
+	err = got_packidx_init_hdr(client->packidx, 1, client->pack.filesize);
 	if (err)
 		return err;
 
@@ -1403,9 +1416,10 @@ verify_packfile(void)
 		if (ref_update->delete_ref)
 			continue;
 
-		RB_FOREACH(pe, got_pathlist_head, repo_write.protected_tag_namespaces) {
+		RB_FOREACH(pe, got_pathlist_head,
+		    &repo_write.protected_tag_namespaces) {
 			err = protect_tag_namespace(pe->path, &client->pack,
-			    packidx, ref_update);
+			    client->packidx, ref_update);
 			if (err)
 				goto done;
 		}
@@ -1423,7 +1437,7 @@ verify_packfile(void)
 			got_object_close(obj);
 			obj = NULL;
 		} else {
-			int idx = got_packidx_get_object_idx(packidx,
+			int idx = got_packidx_get_object_idx(client->packidx,
 			    &ref_update->new_id);
 			if (idx == -1) {
 				got_object_id_hex(&ref_update->new_id,
@@ -1435,26 +1449,23 @@ verify_packfile(void)
 			}
 		}
 
-		RB_FOREACH(pe, got_pathlist_head, repo_write.protected_branch_namespaces)
-		{
+		RB_FOREACH(pe, got_pathlist_head,
+		    &repo_write.protected_branch_namespaces) {
 			err = protect_branch_namespace(pe->path,
-			    &client->pack, packidx, ref_update);
+			    &client->pack, client->packidx, ref_update);
 			if (err)
 				goto done;
 		}
-		RB_FOREACH(pe, got_pathlist_head, repo_write.protected_branches)
-		{
+		RB_FOREACH(pe, got_pathlist_head,
+		    &repo_write.protected_branches) {
 			err = protect_branch(pe->path, &client->pack,
-			    packidx, ref_update);
+			    client->packidx, ref_update);
 			if (err)
 				goto done;
 		}
 	}
 
 done:
-	close_err = got_packidx_close(packidx);
-	if (close_err && err == NULL)
-		err = close_err;
 	free(id_str);
 	if (obj)
 		got_object_close(obj);
@@ -1477,21 +1488,21 @@ protect_refs_from_deletion(void)
 		refname = got_ref_get_name(ref_update->ref);
 
 		RB_FOREACH(pe, got_pathlist_head,
-		    repo_write.protected_tag_namespaces) {
+		    &repo_write.protected_tag_namespaces) {
 			err = protect_ref_namespace(refname, pe->path);
 			if (err)
 				return err;
 		}
 
 		RB_FOREACH(pe, got_pathlist_head,
-		    repo_write.protected_branch_namespaces) {
+		    &repo_write.protected_branch_namespaces) {
 			err = protect_ref_namespace(refname, pe->path);
 			if (err)
 				return err;
 		}
 
-		RB_FOREACH(pe, got_pathlist_head, repo_write.protected_branches)
-		{
+		RB_FOREACH(pe, got_pathlist_head,
+		    &repo_write.protected_branches) {
 			if (strcmp(refname, pe->path) == 0) {
 				return got_error_fmt(GOT_ERR_REF_PROTECTED,
 				    "%s", refname);
@@ -1502,6 +1513,60 @@ protect_refs_from_deletion(void)
 	return NULL;
 }
 
+static const struct got_error *
+protect_refs_from_moving(void)
+{
+	const struct got_error *err = NULL;
+	struct repo_write_client *client = &repo_write_client;
+	struct gotd_ref_update *ref_update;
+	struct got_pathlist_entry *pe;
+	const char *refname;
+
+	STAILQ_FOREACH(ref_update, &client->ref_updates, entry) {
+		if (ref_update->delete_ref)
+			continue;
+
+		if (got_object_id_cmp(&ref_update->old_id,
+		    &ref_update->new_id) == 0)
+			continue;
+
+		refname = got_ref_get_name(ref_update->ref);
+
+		RB_FOREACH(pe, got_pathlist_head,
+		    &repo_write.protected_tag_namespaces) {
+			err = protect_ref_namespace(refname, pe->path);
+			if (err)
+				return err;
+		}
+
+		RB_FOREACH(pe, got_pathlist_head,
+		    &repo_write.protected_branch_namespaces) {
+			if (strcmp(pe->path, refname) != 0)
+				continue;
+
+			err = protect_require_yca(&ref_update->new_id, 0,
+			    NULL, NULL, ref_update->ref);
+			if (err)
+				return err;
+			break;
+		}
+
+		RB_FOREACH(pe, got_pathlist_head,
+		    &repo_write.protected_branches) {
+			if (strcmp(pe->path, refname) != 0)
+				continue;
+
+			err = protect_require_yca(&ref_update->new_id, 0,
+			    NULL, NULL, ref_update->ref);
+			if (err)
+				return err;
+			break;
+		}
+	}
+
+	return NULL;
+}
+
 static const struct got_error *
 install_packfile(struct gotd_imsgev *iev)
 {
@@ -1513,7 +1578,7 @@ install_packfile(struct gotd_imsgev *iev)
 	memcpy(inst.pack_sha1, client->pack_sha1, SHA1_DIGEST_LENGTH);
 
 	ret = gotd_imsg_compose_event(iev, GOTD_IMSG_PACKFILE_INSTALL,
-	    PROC_REPO_WRITE, -1, &inst, sizeof(inst));
+	    GOTD_PROC_REPO_WRITE, -1, &inst, sizeof(inst));
 	if (ret == -1)
 		return got_error_from_errno("imsg_compose PACKFILE_INSTALL");
 
@@ -1530,7 +1595,7 @@ send_ref_updates_start(int nref_updates, struct gotd_imsgev *iev)
 	istart.nref_updates = nref_updates;
 
 	ret = gotd_imsg_compose_event(iev, GOTD_IMSG_REF_UPDATES_START,
-	    PROC_REPO_WRITE, -1, &istart, sizeof(istart));
+	    GOTD_PROC_REPO_WRITE, -1, &istart, sizeof(istart));
 	if (ret == -1)
 		return got_error_from_errno("imsg_compose REF_UPDATES_START");
 
@@ -1554,8 +1619,8 @@ send_ref_update(struct gotd_ref_update *ref_update, struct gotd_imsgev *iev)
 	iref.name_len = strlen(refname);
 
 	len = sizeof(iref) + iref.name_len;
-	wbuf = imsg_create(&iev->ibuf, GOTD_IMSG_REF_UPDATE, PROC_REPO_WRITE,
-	    repo_write.pid, len);
+	wbuf = imsg_create(&iev->ibuf, GOTD_IMSG_REF_UPDATE,
+	    GOTD_PROC_REPO_WRITE, repo_write.pid, len);
 	if (wbuf == NULL)
 		return got_error_from_errno("imsg_create REF_UPDATE");
 
@@ -1727,8 +1792,8 @@ print_diffstat(struct got_diffstat_cb_arg *dsa, int fd)
 		int pad = dsa->max_path_len - pe->path_len + 1;
 
 		dprintf(fd, " %c  %s%*c | %*d+ %*d-\n", cp->status,
-		     pe->path, pad, ' ', dsa->add_cols + 1, cp->add,
-		     dsa->rm_cols + 1, cp->rm);
+		    pe->path, pad, ' ', dsa->add_cols + 1, cp->add,
+		    dsa->rm_cols + 1, cp->rm);
 	}
 	dprintf(fd,
 	    "\n%d file%s changed, %d insertion%s(+), %d deletion%s(-)\n\n",
@@ -2135,7 +2200,7 @@ render_notification(struct imsg *imsg, struct gotd_imsgev *iev)
 	}
 
 	len = sizeof(ireq) + ireq.refname_len;
-	wbuf = imsg_create(&iev->ibuf, GOTD_IMSG_NOTIFY, PROC_REPO_WRITE,
+	wbuf = imsg_create(&iev->ibuf, GOTD_IMSG_NOTIFY, GOTD_PROC_REPO_WRITE,
 	    repo_write.pid, len);
 	if (wbuf == NULL) {
 		err = got_error_from_errno("imsg_create REF");
@@ -2159,6 +2224,345 @@ done:
 	return err;
 }
 
+static const struct got_error *
+send_packfile_received(struct gotd_imsgev *iev, int have_packfile)
+{
+	struct gotd_imsg_packfile_received recvd;
+	int ret;
+
+	if (have_packfile)
+		recvd.pack_empty = 0;
+	else
+		recvd.pack_empty = 1;
+
+	ret = gotd_imsg_compose_event(iev, GOTD_IMSG_PACKFILE_RECEIVED,
+	    GOTD_PROC_REPO_WRITE, -1, &recvd, sizeof(recvd));
+	if (ret == -1)
+		return got_error_from_errno("imsg_compose PACKFILE_RECEIVED");
+
+	return NULL;
+}
+
+static const struct got_error *
+open_tree(struct got_tree_object **tree, struct got_pack *pack,
+    struct got_packidx *packidx, int idx, struct got_object_id *tree_id)
+{
+	const struct got_error *err;
+	struct got_parsed_tree_entry *entries = NULL;
+	size_t nentries = 0, nentries_alloc = 0, i;
+	struct got_object *obj = NULL;
+	uint8_t *buf = NULL;
+	size_t len;
+
+	*tree = NULL;
+
+	err = got_packfile_open_object(&obj, pack, packidx, idx, tree_id);
+	if (err)
+		return err;
+
+	if (obj->type != GOT_OBJ_TYPE_TREE) {
+		err = got_error(GOT_ERR_OBJ_TYPE);
+		goto done;
+	}
+
+	err = got_packfile_extract_object_to_mem(&buf, &len, obj, pack);
+	if (err)
+		goto done;
+	
+	err = got_object_parse_tree(&entries, &nentries, &nentries_alloc,
+	    buf, len, GOT_HASH_SHA1);
+	if (err)
+		goto done;
+
+	*tree = malloc(sizeof(**tree));
+	if (*tree == NULL) {
+		err = got_error_from_errno("malloc");
+		goto done;
+	}
+	(*tree)->entries = calloc(nentries, sizeof(struct got_tree_entry));
+	if ((*tree)->entries == NULL) {
+		err = got_error_from_errno("malloc");
+		goto done;
+	}
+	(*tree)->nentries = nentries;
+	(*tree)->refcnt = 0;
+
+	for (i = 0; i < nentries; i++) {
+		struct got_parsed_tree_entry *pe = &entries[i];
+		struct got_tree_entry *te = &(*tree)->entries[i];
+
+		if (strlcpy(te->name, pe->name,
+		    sizeof(te->name)) >= sizeof(te->name)) {
+			err = got_error(GOT_ERR_NO_SPACE);
+			goto done;
+		}
+		memcpy(te->id.hash, pe->id, pe->digest_len);
+		te->id.algo = GOT_HASH_SHA1;
+		te->mode = pe->mode;
+		te->idx = i;
+	}
+done:
+	if (err && *tree) {
+		got_object_tree_close(*tree);
+		*tree = NULL;
+	}
+	got_object_close(obj);
+	free(buf);
+	return err;
+}
+
+static struct got_tree_entry *
+find_entry_by_name(struct got_tree_object *tree, const char *name, size_t len)
+{
+	int i;
+
+	/* Note that tree entries are sorted in strncmp() order. */
+	for (i = 0; i < tree->nentries; i++) {
+		struct got_tree_entry *te = &tree->entries[i];
+		int cmp = strncmp(te->name, name, len);
+		if (cmp < 0)
+			continue;
+		if (cmp > 0)
+			break;
+		if (te->name[len] == '\0')
+			return te;
+	}
+	return NULL;
+}
+
+static const struct got_error *
+find_id_by_path(struct got_object_id **id,
+    struct got_pack *pack, struct got_packidx *packidx,
+    struct got_tree_object *tree, const char *path)
+{
+	const struct got_error *err = NULL;
+	struct got_tree_object *subtree = NULL;
+	struct got_tree_entry *te = NULL;
+	const char *seg, *s;
+	size_t seglen;
+
+	*id = NULL;
+
+	s = path;
+	while (s[0] == '/')
+		s++;
+	seg = s;
+	seglen = 0;
+	subtree = tree;
+	while (*s) {
+		struct got_tree_object *next_tree;
+
+		if (*s != '/') {
+			s++;
+			seglen++;
+			if (*s)
+				continue;
+		}
+
+		te = find_entry_by_name(subtree, seg, seglen);
+		if (te == NULL)
+			break;
+
+		if (*s == '\0')
+			break;
+
+		seg = s + 1;
+		seglen = 0;
+		s++;
+		if (*s) {
+			int idx;
+
+			idx = got_packidx_get_object_idx(packidx, &te->id);
+			if (idx == -1)  {
+				te = NULL;
+				break;
+			}
+			err = open_tree(&next_tree, pack, packidx, idx,
+			    &te->id);
+			te = NULL;
+			if (err)
+				goto done;
+			if (subtree != tree)
+				got_object_tree_close(subtree);
+			subtree = next_tree;
+		}
+	}
+
+	if (te) {
+		*id = got_object_id_dup(&te->id);
+		if (*id == NULL)
+			return got_error_from_errno("got_object_id_dup");
+	} else
+		err = got_error_path(path, GOT_ERR_NO_TREE_ENTRY);
+done:
+	if (subtree && subtree != tree)
+		got_object_tree_close(subtree);
+	return err;
+}
+
+static const struct got_error *
+get_content_from_packfile(struct gotd_imsgev *iev, struct imsg *imsg)
+{
+	const struct got_error *err = NULL;
+	struct repo_write_client *client = &repo_write_client;
+	struct gotd_imsg_packfile_get_content content_req;
+	struct gotd_imsg_packfile_content_written written_resp;
+	size_t datalen;
+	char *refname = NULL;
+	char *path = NULL;
+	int fd = -1, idx;
+	struct gotd_ref_update *ref_update;
+	struct got_object_id *tree_id = NULL, *content_id = NULL;
+	struct got_object *obj = NULL;
+	uint8_t *buf = NULL;
+	size_t len;
+	struct got_commit_object *commit = NULL;
+	struct got_tree_object *tree = NULL;
+	FILE *outfile = NULL;
+
+	memset(&written_resp, 0, sizeof(written_resp));
+
+	datalen = imsg->hdr.len - IMSG_HEADER_SIZE;
+	if (datalen < sizeof(content_req))
+		return got_error(GOT_ERR_PRIVSEP_LEN);
+	memcpy(&content_req, imsg->data, sizeof(content_req)); 
+
+	if (content_req.path_len == 0 ||
+	    content_req.refname_len == 0 ||
+	    datalen != sizeof(content_req) + content_req.refname_len +
+	    content_req.path_len)
+		return got_error(GOT_ERR_PRIVSEP_LEN);
+	
+	refname = strndup(imsg->data + sizeof(content_req),
+	    content_req.refname_len);
+	if (strlen(refname) != content_req.refname_len) {
+		err = got_error(GOT_ERR_PRIVSEP_LEN);
+		goto done;
+	}
+
+	path = strndup(imsg->data + sizeof(content_req) +
+	    content_req.refname_len, content_req.path_len);
+	if (strlen(path) != content_req.path_len) {
+		err = got_error(GOT_ERR_PRIVSEP_LEN);
+		goto done;
+	}
+
+	if (got_path_is_root_dir(path)) {
+		err = got_error(GOT_ERR_PRIVSEP_MSG);
+		goto done;
+	}
+
+	fd = imsg_get_fd(imsg);
+	if (fd == -1) {
+		err = got_error(GOT_ERR_PRIVSEP_NO_FD);
+		goto done;
+	}
+
+	outfile = fdopen(fd, "w");
+	if (outfile == NULL) {
+		err = got_error_from_errno("fdopen");
+		goto done;
+	}
+	fd = -1;
+
+	STAILQ_FOREACH(ref_update, &client->ref_updates, entry) {
+		if (strcmp(got_ref_get_name(ref_update->ref), refname) == 0)
+			break;
+	}
+	if (ref_update == NULL)
+		goto send_response;
+
+	written_resp.ref_found = 1;	
+
+	idx = got_packidx_get_object_idx(client->packidx, &ref_update->new_id);
+	if (idx == -1) 
+		goto send_response;
+
+	err = got_packfile_open_object(&obj, &client->pack, client->packidx,
+	   idx, &ref_update->new_id);
+	if (err)
+		goto done;
+
+	if (obj->type != GOT_OBJ_TYPE_COMMIT)
+		goto send_response;
+
+	err = got_packfile_extract_object_to_mem(&buf, &len, obj,
+	    &client->pack);
+	if (err)
+		goto done;
+	got_object_close(obj);
+	obj = NULL;
+	
+	err = got_object_parse_commit(&commit, buf, len, GOT_HASH_SHA1);
+	if (err)
+		goto done;
+
+	tree_id = got_object_id_dup(commit->tree_id);
+	if (tree_id == NULL)
+		err = got_error_from_errno("got_object_id_dup");
+
+	got_object_commit_close(commit);
+	commit = NULL;
+	free(buf);
+	buf = NULL;
+	len = 0;
+
+	idx = got_packidx_get_object_idx(client->packidx, tree_id);
+	if (idx == -1) 
+		goto send_response;
+
+	err = open_tree(&tree, &client->pack, client->packidx, idx, tree_id);
+	if (err)
+		goto done;
+
+	err = find_id_by_path(&content_id, &client->pack, client->packidx,
+	    tree, path);
+	if (err)
+		goto done;
+
+	idx = got_packidx_get_object_idx(client->packidx, content_id);
+	if (idx == -1) 
+		goto send_response;
+
+	err = got_packfile_open_object(&obj, &client->pack, client->packidx,
+	   idx, content_id);
+	if (err)
+		goto done;
+
+	if (obj->type != GOT_OBJ_TYPE_BLOB)
+		goto send_response;
+
+	err = got_packfile_extract_object(&client->pack, obj, outfile,
+	    repo_write.base_file, repo_write.accum_file);
+	if (err)
+		goto done;
+	
+	written_resp.wrote_content = 1;	
+send_response:
+	if (imsg_compose(&iev->ibuf, GOTD_IMSG_PACKFILE_CONTENT_WRITTEN,
+	    GOTD_PROC_REPO_WRITE, repo_write.pid, -1, &written_resp,
+	    sizeof(written_resp)) == -1) {
+		err = got_error_from_errno("imsg_compose "
+		    "PACKFILE_CONTENT_WRITTEn");
+		goto done;
+	}
+	err = gotd_imsg_flush(&iev->ibuf);
+done:
+	if (fd != -1 && close(fd) == -1 && err == NULL)
+		err = got_error_from_errno("close");
+	if (outfile && fclose(outfile) == EOF && err == NULL)
+		err = got_error_from_errno("fclose");
+	free(refname);
+	free(path);
+	free(tree_id);
+	free(buf);
+	if (obj)
+		got_object_close(obj);
+	if (commit)
+		got_object_commit_close(commit);
+	return err;
+}
+
 static void
 repo_write_dispatch_session(int fd, short event, void *arg)
 {
@@ -2168,7 +2572,7 @@ repo_write_dispatch_session(int fd, short event, void *arg)
 	struct imsg imsg;
 	struct repo_write_client *client = &repo_write_client;
 	ssize_t n;
-	int shut = 0, have_packfile = 0;
+	int shut = 0;
 
 	if (event & EV_READ) {
 		if ((n = imsgbuf_read(ibuf)) == -1)
@@ -2183,7 +2587,7 @@ repo_write_dispatch_session(int fd, short event, void *arg)
 			fatalx("%s", err->msg);
 	}
 
-	for (;;) {
+	while (err == NULL) {
 		if ((n = imsg_get(ibuf, &imsg)) == -1)
 			fatal("%s: imsg_get error", __func__);
 		if (n == 0)	/* No more messages. */
@@ -2225,18 +2629,42 @@ repo_write_dispatch_session(int fd, short event, void *arg)
 			err = protect_refs_from_deletion();
 			if (err)
 				break;
-			err = recv_packfile(&have_packfile, &imsg);
+			err = recv_packfile(&repo_write.have_packfile, &imsg);
 			if (err) {
 				log_warnx("receive packfile: %s", err->msg);
 				break;
 			}
-			if (have_packfile) {
+			if (repo_write.have_packfile) {
 				err = verify_packfile();
 				if (err) {
 					log_warnx("verify packfile: %s",
 					    err->msg);
 					break;
 				}
+			} else {
+				/*
+				 * Clients sending empty pack files might be
+				 * attempting to move a protected reference.
+				 */
+				err = protect_refs_from_moving();
+				if (err)
+					break;
+			}
+			err = send_packfile_received(iev,
+			    repo_write.have_packfile);
+			if (err)
+				log_warnx("receive packfile: %s", err->msg);
+			break;
+		case GOTD_IMSG_PACKFILE_GET_CONTENT:
+			if (!repo_write.have_packfile ||
+			    client->packidx == NULL) {
+				err = got_error(GOT_ERR_PRIVSEP_MSG);
+				break;
+			}
+			err = get_content_from_packfile(iev, &imsg);
+			break;
+		case GOTD_IMSG_PACKFILE_VERIFIED:
+			if (repo_write.have_packfile) {
 				err = install_packfile(iev);
 				if (err) {
 					log_warnx("install packfile: %s",
@@ -2251,9 +2679,8 @@ repo_write_dispatch_session(int fd, short event, void *arg)
 				repo_write.repo->pack_path_mtime.tv_nsec = 0;
 			}
 			err = update_refs(iev);
-			if (err) {
+			if (err)
 				log_warnx("update refs: %s", err->msg);
-			}
 			break;
 		case GOTD_IMSG_NOTIFY:
 			err = render_notification(&imsg, iev);
@@ -2264,6 +2691,7 @@ repo_write_dispatch_session(int fd, short event, void *arg)
 			break;
 		default:
 			log_debug("unexpected imsg %d", imsg.hdr.type);
+			err = got_error(GOT_ERR_PRIVSEP_MSG);
 			break;
 		}
 
@@ -2271,9 +2699,8 @@ repo_write_dispatch_session(int fd, short event, void *arg)
 	}
 
 	if (!shut && check_cancelled(NULL) == NULL) {
-		if (err &&
-		    gotd_imsg_send_error_event(iev, PROC_REPO_WRITE,
-		        client->id, err) == -1) {
+		if (err && gotd_imsg_send_error_event(iev,
+		    GOTD_PROC_REPO_WRITE, client->id, err) == -1) {
 			log_warnx("could not send error to parent: %s",
 			    err->msg);
 		}
@@ -2325,6 +2752,7 @@ repo_write_dispatch(int fd, short event, void *arg)
 	ssize_t n;
 	int shut = 0;
 	struct repo_write_client *client = &repo_write_client;
+	size_t npaths;
 
 	if (event & EV_READ) {
 		if ((n = imsgbuf_read(ibuf)) == -1)
@@ -2351,6 +2779,68 @@ repo_write_dispatch(int fd, short event, void *arg)
 			break;
 
 		switch (imsg.hdr.type) {
+		case GOTD_IMSG_PROTECTED_TAG_NAMESPACES:
+			if (repo_write.protected_refs_cur != NULL ||
+			    repo_write.nprotected_refs_needed != 0) {
+				err = got_error(GOT_ERR_PRIVSEP_MSG);
+				break;
+			}
+			err = gotd_imsg_recv_pathlist(&npaths, &imsg);
+			if (err)
+				break;
+			repo_write.protected_refs_cur =
+			    &repo_write.protected_tag_namespaces;
+			repo_write.nprotected_refs_needed = npaths;
+			repo_write.nprotected_refs_received = 0;
+			break;
+		case GOTD_IMSG_PROTECTED_BRANCH_NAMESPACES:
+			if (repo_write.protected_refs_cur != NULL ||
+			    repo_write.nprotected_refs_needed != 0) {
+				err = got_error(GOT_ERR_PRIVSEP_MSG);
+				break;
+			}
+			err = gotd_imsg_recv_pathlist(&npaths, &imsg);
+			if (err)
+				break;
+			repo_write.protected_refs_cur =
+			    &repo_write.protected_branch_namespaces;
+			repo_write.nprotected_refs_needed = npaths;
+			repo_write.nprotected_refs_received = 0;
+			break;
+		case GOTD_IMSG_PROTECTED_BRANCHES:
+			if (repo_write.protected_refs_cur != NULL ||
+			    repo_write.nprotected_refs_needed != 0) {
+				err = got_error(GOT_ERR_PRIVSEP_MSG);
+				break;
+			}
+			err = gotd_imsg_recv_pathlist(&npaths, &imsg);
+			if (err)
+				break;
+			repo_write.protected_refs_cur =
+			    &repo_write.protected_branches;
+			repo_write.nprotected_refs_needed = npaths;
+			repo_write.nprotected_refs_received = 0;
+			break;
+		case GOTD_IMSG_PROTECTED_TAG_NAMESPACES_ELEM:
+		case GOTD_IMSG_PROTECTED_BRANCH_NAMESPACES_ELEM:
+		case GOTD_IMSG_PROTECTED_BRANCHES_ELEM:
+			if (repo_write.protected_refs_cur == NULL ||
+			    repo_write.nprotected_refs_needed == 0 ||
+			    repo_write.nprotected_refs_received >=
+			    repo_write.nprotected_refs_needed) {
+				err = got_error(GOT_ERR_PRIVSEP_MSG);
+				break;
+			}
+			err = gotd_imsg_recv_pathlist_elem(&imsg,
+			    repo_write.protected_refs_cur);
+			if (err)
+				break;
+			if (++repo_write.nprotected_refs_received >=
+			    repo_write.nprotected_refs_needed) {
+				repo_write.protected_refs_cur = NULL;
+				repo_write.nprotected_refs_needed = 0;
+			}
+			break;
 		case GOTD_IMSG_CONNECT_REPO_CHILD:
 			err = recv_connect(&imsg);
 			break;
@@ -2363,7 +2853,7 @@ repo_write_dispatch(int fd, short event, void *arg)
 		imsg_free(&imsg);
 	}
 
-	if (err && gotd_imsg_send_error_event(iev, PROC_REPO_WRITE,
+	if (err && gotd_imsg_send_error_event(iev, GOTD_PROC_REPO_WRITE,
 	    client->id, err) == -1)
 		log_warnx("could not send error to parent: %s", err->msg);
 done:
@@ -2378,11 +2868,8 @@ done:
 
 void
 repo_write_main(const char *title, const char *repo_path,
-    int *pack_fds, int *temp_fds,
-    FILE *diff_f1, FILE *diff_f2, int diff_fd1, int diff_fd2,
-    struct got_pathlist_head *protected_tag_namespaces,
-    struct got_pathlist_head *protected_branch_namespaces,
-    struct got_pathlist_head *protected_branches)
+    int *pack_fds, int *temp_fds, FILE *base_file, FILE *accum_file,
+    FILE *diff_f1, FILE *diff_f2, int diff_fd1, int diff_fd2)
 {
 	const struct got_error *err = NULL;
 	struct repo_write_client *client = &repo_write_client;
@@ -2397,11 +2884,13 @@ repo_write_main(const char *title, const char *repo_path,
 	repo_write.pid = getpid();
 	repo_write.pack_fds = pack_fds;
 	repo_write.temp_fds = temp_fds;
+	repo_write.base_file = base_file;
+	repo_write.accum_file = accum_file;
 	repo_write.session_fd = -1;
 	repo_write.session_iev.ibuf.fd = -1;
-	repo_write.protected_tag_namespaces = protected_tag_namespaces;
-	repo_write.protected_branch_namespaces = protected_branch_namespaces;
-	repo_write.protected_branches = protected_branches;
+	RB_INIT(&repo_write.protected_tag_namespaces);
+	RB_INIT(&repo_write.protected_branch_namespaces);
+	RB_INIT(&repo_write.protected_branches);
 	repo_write.diff.f1 = diff_f1;
 	repo_write.diff.f2 = diff_f2;
 	repo_write.diff.fd1 = diff_fd1;
@@ -2440,7 +2929,7 @@ repo_write_main(const char *title, const char *repo_path,
 	iev.handler_arg = NULL;
 	event_set(&iev.ev, iev.ibuf.fd, EV_READ, repo_write_dispatch, &iev);
 	if (gotd_imsg_compose_event(&iev, GOTD_IMSG_REPO_CHILD_READY,
-	    PROC_REPO_WRITE, -1, NULL, 0) == -1) {
+	    GOTD_PROC_REPO_WRITE, -1, NULL, 0) == -1) {
 		err = got_error_from_errno("imsg compose REPO_CHILD_READY");
 		goto done;
 	}
@@ -2475,6 +2964,8 @@ repo_write_shutdown(void)
 		free(ref_update);
 	}
 
+	if (client->packidx)
+		got_packidx_close(client->packidx);
 	got_pack_close(&client->pack);
 	if (client->fd != -1)
 		close(client->fd);
