@@ -43,63 +43,139 @@
 #include <netdb.h>
 #include <poll.h>
 #include <pwd.h>
+#include <siphash.h>
 #include <stddef.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
 
-#include "got_error.h"
-#include "got_opentemp.h"
 #include "got_reference.h"
-#include "got_repository.h"
-#include "got_privsep.h"
 
 #include "gotwebd.h"
+#include "log.h"
 #include "tmpl.h"
 
 #define SOCKS_BACKLOG 5
 #define MAXIMUM(a, b)	(((a) > (b)) ? (a) : (b))
 
-volatile int client_cnt;
+static volatile int client_cnt;
 
 static struct timeval	timeout = { TIMEOUT_DEFAULT, 0 };
 
 static void	 sockets_sighdlr(int, short, void *);
 static void	 sockets_shutdown(void);
-static void	 sockets_launch(void);
-static void	 sockets_purge(struct gotwebd *);
+static void	 sockets_launch(struct gotwebd *);
 static void	 sockets_accept_paused(int, short, void *);
-static void	 sockets_rlimit(int);
 
 static void	 sockets_dispatch_main(int, short, void *);
-static int	 sockets_unix_socket_listen(struct gotwebd *, struct socket *);
+static int	 sockets_unix_socket_listen(struct gotwebd *, struct socket *, uid_t, gid_t);
 static int	 sockets_create_socket(struct address *);
 static int	 sockets_accept_reserve(int, struct sockaddr *, socklen_t *,
 		    int, volatile int *);
 
-static struct socket *sockets_conf_new_socket_unix(struct gotwebd *,
-		    struct server *, int);
-static struct socket *sockets_conf_new_socket_fcgi(struct gotwebd *,
-		    struct server *, int, struct address *);
+static struct socket *sockets_conf_new_socket(struct gotwebd *,
+		    int, struct address *);
 
 int cgi_inflight = 0;
 
+/* Request hash table needs some spare room to avoid collisions. */
+struct requestlist requests[GOTWEBD_MAXCLIENTS * 4];
+static SIPHASH_KEY requests_hash_key;
+
+static void
+requests_init(void)
+{
+	int i;
+
+	arc4random_buf(&requests_hash_key, sizeof(requests_hash_key));
+
+	for (i = 0; i < nitems(requests); i++)
+		TAILQ_INIT(&requests[i]);
+}
+
+static uint64_t
+request_hash(uint32_t request_id)
+{
+	return SipHash24(&requests_hash_key, &request_id, sizeof(request_id));
+}
+
+static void
+add_request(struct request *c)
+{
+	uint64_t slot = request_hash(c->request_id) % nitems(requests);
+	TAILQ_INSERT_HEAD(&requests[slot], c, entry);
+	client_cnt++;
+}
+
+void
+sockets_del_request(struct request *c)
+{
+	uint64_t slot = request_hash(c->request_id) % nitems(requests);
+	TAILQ_REMOVE(&requests[slot], c, entry);
+	client_cnt--;
+}
+
+static struct request *
+find_request(uint32_t request_id)
+{
+	uint64_t slot;
+	struct request *c;
+
+	slot = request_hash(request_id) % nitems(requests);
+	TAILQ_FOREACH(c, &requests[slot], entry) {
+		if (c->request_id == request_id)
+			return c;
+	}
+
+	return NULL;
+}
+
+static void
+requests_purge(void)
+{
+	uint64_t slot;
+	struct request *c;
+
+	for (slot = 0; slot < nitems(requests); slot++) {
+		while (!TAILQ_EMPTY(&requests[slot])) {
+			c = TAILQ_FIRST(&requests[slot]);
+			fcgi_cleanup_request(c);
+		}
+	}
+}
+
+static uint32_t
+get_request_id(void)
+{
+	int duplicate = 0;
+	uint32_t id;
+
+	do {
+		id = arc4random();
+		duplicate = (find_request(id) != NULL);
+	} while (duplicate || id == 0);
+
+	return id;
+}
+
 void
 sockets(struct gotwebd *env, int fd)
 {
 	struct event	 sighup, sigint, sigusr1, sigchld, sigterm;
+	struct event_base *evb;
 
-	event_init();
+	requests_init();
+
+	evb = event_init();
 
 	sockets_rlimit(-1);
 
-	if (config_init(env) == -1)
-		fatal("failed to initialize configuration");
-
 	if ((env->iev_parent = malloc(sizeof(*env->iev_parent))) == NULL)
 		fatal("malloc");
-	imsg_init(&env->iev_parent->ibuf, fd);
+	if (imsgbuf_init(&env->iev_parent->ibuf, fd) == -1)
+		fatal("imsgbuf_init");
+	imsgbuf_allow_fdpass(&env->iev_parent->ibuf);
 	env->iev_parent->handler = sockets_dispatch_main;
 	env->iev_parent->data = env->iev_parent;
 	event_set(&env->iev_parent->ev, fd, EV_READ, sockets_dispatch_main,
@@ -120,100 +196,37 @@ sockets(struct gotwebd *env, int fd)
 	signal_add(&sigterm, NULL);
 
 #ifndef PROFILE
-	if (pledge("stdio rpath inet recvfd proc exec sendfd unveil",
-	    NULL) == -1)
+	if (pledge("stdio inet recvfd sendfd", NULL) == -1)
 		fatal("pledge");
 #endif
 
 	event_dispatch();
+	event_base_free(evb);
 	sockets_shutdown();
 }
 
 void
 sockets_parse_sockets(struct gotwebd *env)
 {
-	struct server *srv;
 	struct address *a;
 	struct socket *new_sock = NULL;
 	int sock_id = 1;
 
-	TAILQ_FOREACH(srv, &env->servers, entry) {
-		if (srv->unix_socket) {
-			new_sock = sockets_conf_new_socket_unix(env, srv,
-			    sock_id);
-			if (new_sock) {
-				sock_id++;
-				TAILQ_INSERT_TAIL(&env->sockets, new_sock,
-				    entry);
-			}
-		}
-
-		if (srv->fcgi_socket) {
-			if (TAILQ_EMPTY(&srv->al)) {
-				fatalx("%s: server %s has no IP addresses to "
-				    "listen for FCGI connections", __func__,
-				    srv->name);
-			}
-			TAILQ_FOREACH(a, &srv->al, entry) {
-				if (a->ss.ss_family != AF_INET &&
-				    a->ss.ss_family != AF_INET6)
-					continue;
-				new_sock = sockets_conf_new_socket_fcgi(env,
-				    srv, sock_id, a);
-				if (new_sock) {
-					sock_id++;
-					TAILQ_INSERT_TAIL(&env->sockets,
-					    new_sock, entry);
-				}
-			}
+	TAILQ_FOREACH(a, &env->addresses, entry) {
+		new_sock = sockets_conf_new_socket(env, sock_id, a);
+		if (new_sock) {
+			sock_id++;
+			TAILQ_INSERT_TAIL(&env->sockets,
+			    new_sock, entry);
 		}
 	}
 }
 
 static struct socket *
-sockets_conf_new_socket_unix(struct gotwebd *env, struct server *srv, int id)
-{
-	struct socket *sock;
-	int n;
-
-	if ((sock = calloc(1, sizeof(*sock))) == NULL)
-		fatalx("%s: calloc", __func__);
-
-	sock->conf.id = id;
-	sock->fd = -1;
-	sock->conf.af_type = AF_UNIX;
-
-	if (strlcpy(sock->conf.unix_socket_name,
-	    srv->unix_socket_name,
-	    sizeof(sock->conf.unix_socket_name)) >=
-	    sizeof(sock->conf.unix_socket_name)) {
-		free(sock);
-		fatalx("%s: strlcpy", __func__);
-	}
-
-	n = snprintf(sock->conf.name, GOTWEBD_MAXTEXT, "%s_parent",
-	    srv->name);
-	if (n < 0 || (size_t)n >= GOTWEBD_MAXTEXT) {
-		free(sock);
-		fatalx("%s: snprintf", __func__);
-	}
-
-	if (strlcpy(sock->conf.srv_name, srv->name,
-	    sizeof(sock->conf.srv_name)) >= sizeof(sock->conf.srv_name)) {
-		free(sock);
-		fatalx("%s: strlcpy", __func__);
-	}
-
-	return sock;
-}
-
-static struct socket *
-sockets_conf_new_socket_fcgi(struct gotwebd *env, struct server *srv, int id,
-    struct address *a)
+sockets_conf_new_socket(struct gotwebd *env, int id, struct address *a)
 {
 	struct socket *sock;
 	struct address *acp;
-	int n;
 
 	if ((sock = calloc(1, sizeof(*sock))) == NULL)
 		fatalx("%s: calloc", __func__);
@@ -222,21 +235,18 @@ sockets_conf_new_socket_fcgi(struct gotwebd *env, struct server *srv, int id,
 	sock->fd = -1;
 	sock->conf.af_type = a->ss.ss_family;
 
+	if (a->ss.ss_family == AF_UNIX) {
+		struct sockaddr_un *sun;
+
+		sun = (struct sockaddr_un *)&a->ss;
+		if (strlcpy(sock->conf.unix_socket_name, sun->sun_path,
+		    sizeof(sock->conf.unix_socket_name)) >=
+		    sizeof(sock->conf.unix_socket_name))
+			fatalx("unix socket path too long: %s", sun->sun_path);
+	}
+
 	sock->conf.fcgi_socket_port = a->port;
 
-	n = snprintf(sock->conf.name, GOTWEBD_MAXTEXT, "%s_parent",
-	    srv->name);
-	if (n < 0 || (size_t)n >= GOTWEBD_MAXTEXT) {
-		free(sock);
-		fatalx("%s: snprintf", __func__);
-	}
-
-	if (strlcpy(sock->conf.srv_name, srv->name,
-	    sizeof(sock->conf.srv_name)) >= sizeof(sock->conf.srv_name)) {
-		free(sock);
-		fatalx("%s: strlcpy", __func__);
-	}
-
 	acp = &sock->conf.addr;
 
 	memcpy(&acp->ss, &a->ss, sizeof(acp->ss));
@@ -257,14 +267,15 @@ sockets_conf_new_socket_fcgi(struct gotwebd *env, struct server *srv, int id,
 }
 
 static void
-sockets_launch(void)
+sockets_launch(struct gotwebd *env)
 {
 	struct socket *sock;
-	struct server *srv;
-	const struct got_error *error;
+
+	if (env->iev_gotweb == NULL)
+		fatal("gotweb process not connected");
 
 	TAILQ_FOREACH(sock, &gotwebd_env->sockets, entry) {
-		log_debug("%s: configuring socket %d (%d)", __func__,
+		log_info("%s: configuring socket %d (%d)", __func__,
 		    sock->conf.id, sock->fd);
 
 		event_set(&sock->ev, sock->fd, EV_READ | EV_PERSIST,
@@ -275,24 +286,19 @@ sockets_launch(void)
 
 		evtimer_set(&sock->pause, sockets_accept_paused, sock);
 
-		log_debug("%s: running socket listener %d", __func__,
+		log_info("%s: running socket listener %d", __func__,
 		    sock->conf.id);
 	}
 
-	TAILQ_FOREACH(srv, &gotwebd_env->servers, entry) {
-		if (unveil(srv->repos_path, "r") == -1)
-			fatal("unveil %s", srv->repos_path);
-	}
+#ifndef PROFILE
+	if (pledge("stdio inet sendfd", NULL) == -1)
+		fatal("pledge");
+#endif
+	event_add(&env->iev_gotweb->ev, NULL);
 
-	error = got_privsep_unveil_exec_helpers();
-	if (error)
-		fatal("%s", error->msg);
-
-	if (unveil(NULL, NULL) == -1)
-		fatal("unveil");
 }
 
-static void
+void
 sockets_purge(struct gotwebd *env)
 {
 	struct socket *sock, *tsock;
@@ -308,9 +314,116 @@ sockets_purge(struct gotwebd *env)
 		if (sock->fd != -1)
 			close(sock->fd);
 		TAILQ_REMOVE(&env->sockets, sock, entry);
+		free(sock);
 	}
 }
 
+static void
+request_done(struct imsg *imsg)
+{
+	struct request *c;
+	uint32_t request_id;
+	size_t datalen = imsg->hdr.len - IMSG_HEADER_SIZE;
+
+	if (datalen != sizeof(request_id)) {
+		log_warn("IMSG_REQ_DONE with bad data length");
+		return;
+	}
+
+	memcpy(&request_id, imsg->data, sizeof(request_id));
+
+	c = find_request(request_id);
+	if (c == NULL) {
+		log_warnx("no request to clean up found for ID %u",
+		    request_id);
+		return;
+	}
+
+	if (c->client_status == CLIENT_REQUEST)
+		fcgi_create_end_record(c);
+	fcgi_cleanup_request(c);
+}
+
+static void
+server_dispatch_gotweb(int fd, short event, void *arg)
+{
+	struct imsgev		*iev = arg;
+	struct imsgbuf		*ibuf;
+	struct imsg		 imsg;
+	ssize_t			 n;
+	int			 shut = 0;
+
+	ibuf = &iev->ibuf;
+
+	if (event & EV_READ) {
+		if ((n = imsgbuf_read(ibuf)) == -1)
+			fatal("imsgbuf_read error");
+		if (n == 0)	/* Connection closed */
+			shut = 1;
+	}
+	if (event & EV_WRITE) {
+		if (imsgbuf_write(ibuf) == -1)
+			fatal("imsgbuf_write");
+	}
+
+	for (;;) {
+		if ((n = imsg_get(ibuf, &imsg)) == -1)
+			fatal("imsg_get");
+		if (n == 0)	/* No more messages. */
+			break;
+
+		switch (imsg.hdr.type) {
+		case GOTWEBD_IMSG_REQ_DONE:
+			request_done(&imsg);
+			break;
+		default:
+			fatalx("%s: unknown imsg type %d", __func__,
+			    imsg.hdr.type);
+		}
+
+		imsg_free(&imsg);
+	}
+
+	if (!shut)
+		imsg_event_add(iev);
+	else {
+		/* This pipe is dead.  Remove its event handler */
+		event_del(&iev->ev);
+		event_loopexit(NULL);
+	}
+}
+
+static void
+recv_gotweb_pipe(struct gotwebd *env, struct imsg *imsg)
+{
+	struct imsgev *iev;
+	int fd;
+
+	if (env->iev_gotweb != NULL) {
+		log_warn("gotweb pipe already received");
+		return;
+	}
+
+	fd = imsg_get_fd(imsg);
+	if (fd == -1)
+		fatalx("invalid gotweb pipe fd");
+
+	iev = calloc(1, sizeof(*iev));
+	if (iev == NULL)
+		fatal("calloc");
+
+	if (imsgbuf_init(&iev->ibuf, fd) == -1)
+		fatal("imsgbuf_init");
+	imsgbuf_allow_fdpass(&iev->ibuf);
+
+	iev->handler = server_dispatch_gotweb;
+	iev->data = iev;
+	event_set(&iev->ev, fd, EV_READ, server_dispatch_gotweb, iev);
+	imsg_event_add(iev);
+
+	env->iev_gotweb = iev;
+}
+
 static void
 sockets_dispatch_main(int fd, short event, void *arg)
 {
@@ -324,16 +437,14 @@ sockets_dispatch_main(int fd, short event, void *arg)
 	ibuf = &iev->ibuf;
 
 	if (event & EV_READ) {
-		if ((n = imsg_read(ibuf)) == -1 && errno != EAGAIN)
-			fatal("imsg_read error");
+		if ((n = imsgbuf_read(ibuf)) == -1)
+			fatal("imsgbuf_read error");
 		if (n == 0)	/* Connection closed */
 			shut = 1;
 	}
 	if (event & EV_WRITE) {
-		if ((n = msgbuf_write(&ibuf->w)) == -1 && errno != EAGAIN)
-			fatal("msgbuf_write");
-		if (n == 0)	/* Connection closed */
-			shut = 1;
+		if (imsgbuf_write(ibuf) == -1)
+			fatal("imsgbuf_write");
 	}
 
 	for (;;) {
@@ -343,20 +454,20 @@ sockets_dispatch_main(int fd, short event, void *arg)
 			break;
 
 		switch (imsg.hdr.type) {
-		case IMSG_CFG_SRV:
+		case GOTWEBD_IMSG_CFG_SRV:
 			config_getserver(env, &imsg);
 			break;
-		case IMSG_CFG_SOCK:
+		case GOTWEBD_IMSG_CFG_SOCK:
 			config_getsock(env, &imsg);
 			break;
-		case IMSG_CFG_FD:
-			config_getfd(env, &imsg);
-			break;
-		case IMSG_CFG_DONE:
+		case GOTWEBD_IMSG_CFG_DONE:
 			config_getcfg(env, &imsg);
 			break;
-		case IMSG_CTL_START:
-			sockets_launch();
+		case GOTWEBD_IMSG_CTL_PIPE:
+			recv_gotweb_pipe(env, &imsg);
+			break;
+		case GOTWEBD_IMSG_CTL_START:
+			sockets_launch(env);
 			break;
 		default:
 			fatalx("%s: unknown imsg type %d", __func__,
@@ -403,89 +514,77 @@ sockets_sighdlr(int sig, short event, void *arg)
 static void
 sockets_shutdown(void)
 {
-	struct server *srv, *tsrv;
-	struct socket *sock, *tsock;
-
 	sockets_purge(gotwebd_env);
 
-	/* clean sockets */
-	TAILQ_FOREACH_SAFE(sock, &gotwebd_env->sockets, entry, tsock) {
-		TAILQ_REMOVE(&gotwebd_env->sockets, sock, entry);
-		close(sock->fd);
-		free(sock);
+	/* clean servers */
+	while (!TAILQ_EMPTY(&gotwebd_env->servers)) {
+		struct server *srv;
+
+		srv = TAILQ_FIRST(&gotwebd_env->servers);
+		TAILQ_REMOVE(&gotwebd_env->servers, srv, entry);
+		free(srv);
 	}
 
-	/* clean servers */
-	TAILQ_FOREACH_SAFE(srv, &gotwebd_env->servers, entry, tsrv)
-		free(srv);
+	while (!TAILQ_EMPTY(&gotwebd_env->addresses)) {
+		struct address *h;
 
+		h = TAILQ_FIRST(&gotwebd_env->addresses);
+		TAILQ_REMOVE(&gotwebd_env->addresses, h, entry);
+		free(h);
+	}
+
+	requests_purge();
+
+	imsgbuf_clear(&gotwebd_env->iev_parent->ibuf);
+	free(gotwebd_env->iev_parent);
+	imsgbuf_clear(&gotwebd_env->iev_gotweb->ibuf);
+	free(gotwebd_env->iev_gotweb);
 	free(gotwebd_env);
 
 	exit(0);
 }
 
 int
-sockets_privinit(struct gotwebd *env, struct socket *sock)
+sockets_privinit(struct gotwebd *env, struct socket *sock, uid_t uid, gid_t gid)
 {
 	if (sock->conf.af_type == AF_UNIX) {
-		log_debug("%s: initializing unix socket %s", __func__,
+		log_info("%s: initializing unix socket %s", __func__,
 		    sock->conf.unix_socket_name);
-		sock->fd = sockets_unix_socket_listen(env, sock);
-		if (sock->fd == -1) {
-			log_warnx("%s: create unix socket failed", __func__);
+		sock->fd = sockets_unix_socket_listen(env, sock, uid, gid);
+		if (sock->fd == -1)
 			return -1;
-		}
 	}
 
 	if (sock->conf.af_type == AF_INET || sock->conf.af_type == AF_INET6) {
-		log_debug("%s: initializing %s FCGI socket on port %d for %s",
+		log_info("%s: initializing %s FCGI socket on port %d",
 		    __func__, sock->conf.af_type == AF_INET ? "inet" : "inet6",
-		    sock->conf.fcgi_socket_port, sock->conf.name);
+		    sock->conf.fcgi_socket_port);
 		sock->fd = sockets_create_socket(&sock->conf.addr);
-		if (sock->fd == -1) {
-			log_warnx("%s: create FCGI socket failed", __func__);
+		if (sock->fd == -1)
 			return -1;
-		}
 	}
 
 	return 0;
 }
 
 static int
-sockets_unix_socket_listen(struct gotwebd *env, struct socket *sock)
+sockets_unix_socket_listen(struct gotwebd *env, struct socket *sock,
+    uid_t uid, gid_t gid)
 {
-	struct sockaddr_un sun;
-	struct socket *tsock;
 	int u_fd = -1;
 	mode_t old_umask, mode;
+	int flags = SOCK_STREAM | SOCK_NONBLOCK;
 
-	TAILQ_FOREACH(tsock, &env->sockets, entry) {
-		if (strcmp(tsock->conf.unix_socket_name,
-		    sock->conf.unix_socket_name) == 0 &&
-		    tsock->fd != -1)
-			return (tsock->fd);
-	}
-
-	/* TA: FIXME:  this needs upstreaming. */
-	int socket_flags = SOCK_STREAM | SOCK_NONBLOCK;
 #ifdef SOCK_CLOEXEC
-	socket_flags |= SOCK_CLOEXEC;
+	flags |= SOCK_CLOEXEC;
 #endif
-	u_fd = socket(AF_UNIX, socket_flags, 0);
+
+	u_fd = socket(AF_UNIX, flags, 0);
 	if (u_fd == -1) {
 		log_warn("%s: socket", __func__);
 		return -1;
 	}
 
-	sun.sun_family = AF_UNIX;
-	if (strlcpy(sun.sun_path, sock->conf.unix_socket_name,
-	    sizeof(sun.sun_path)) >= sizeof(sun.sun_path)) {
-		log_warn("%s: %s name too long", __func__,
-		    sock->conf.unix_socket_name);
-		close(u_fd);
-		return -1;
-	}
-
 	if (unlink(sock->conf.unix_socket_name) == -1) {
 		if (errno != ENOENT) {
 			log_warn("%s: unlink %s", __func__,
@@ -498,7 +597,8 @@ sockets_unix_socket_listen(struct gotwebd *env, struct socket *sock)
 	old_umask = umask(S_IXUSR|S_IXGRP|S_IWOTH|S_IROTH|S_IXOTH);
 	mode = S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP;
 
-	if (bind(u_fd, (struct sockaddr *)&sun, sizeof(sun)) == -1) {
+	if (bind(u_fd, (struct sockaddr *)&sock->conf.addr.ss,
+	    sock->conf.addr.slen) == -1) {
 		log_warn("%s: bind: %s", __func__, sock->conf.unix_socket_name);
 		close(u_fd);
 		(void)umask(old_umask);
@@ -514,8 +614,7 @@ sockets_unix_socket_listen(struct gotwebd *env, struct socket *sock)
 		return -1;
 	}
 
-	if (chown(sock->conf.unix_socket_name, env->pw->pw_uid,
-	    env->pw->pw_gid) == -1) {
+	if (chown(sock->conf.unix_socket_name, uid, gid) == -1) {
 		log_warn("%s: chown", __func__);
 		close(u_fd);
 		(void)unlink(sock->conf.unix_socket_name);
@@ -553,14 +652,14 @@ sockets_create_socket(struct address *a)
 	flags = fcntl(fd, F_GETFL);
 	flags |= O_NONBLOCK;
 	if (fcntl(fd, F_SETFL, flags) == -1) {
-		log_info("%s: could not enable non-blocking I/O", __func__);
+		log_warn("%s: could not enable non-blocking I/O", __func__);
 		close(fd);
 		return -1;
 	}
 
 	if (bind(fd, (struct sockaddr *)&a->ss, a->slen) == -1) {
 		close(fd);
-		log_info("%s: can't bind to port %d", __func__, a->port);
+		log_warn("%s: can't bind to port %d", __func__, a->port);
 		return -1;
 	}
 
@@ -581,7 +680,7 @@ sockets_accept_reserve(int sockfd, struct sockaddr *addr, socklen_t *addrlen,
 
 	if (getdtablecount() + reserve +
 	    ((*counter + 1) * FD_NEEDED) >= getdtablesize()) {
-		log_debug("inflight fds exceeded");
+		log_warnx("inflight fds exceeded");
 		errno = EMFILE;
 		return -1;
 	}
@@ -615,6 +714,7 @@ sockets_socket_accept(int fd, short event, void *arg)
 	struct sockaddr_storage ss;
 	struct timeval backoff;
 	struct request *c = NULL;
+	uint8_t *buf = NULL;
 	socklen_t len;
 	int s;
 
@@ -638,8 +738,10 @@ sockets_socket_accept(int fd, short event, void *arg)
 			return;
 		case EMFILE:
 		case ENFILE:
+			log_warn("accept");
 			event_del(&sock->ev);
-			evtimer_add(&sock->pause, &backoff);
+			if (!evtimer_pending(&sock->pause, NULL))
+				evtimer_add(&sock->pause, &backoff);
 			return;
 		default:
 			log_warn("%s: accept", __func__);
@@ -651,37 +753,40 @@ sockets_socket_accept(int fd, short event, void *arg)
 
 	c = calloc(1, sizeof(struct request));
 	if (c == NULL) {
-		log_warn("%s", __func__);
+		log_warn("%s: calloc", __func__);
 		close(s);
 		cgi_inflight--;
 		return;
 	}
 
-	c->tp = template(c, &fcgi_write, c->outbuf, sizeof(c->outbuf));
-	if (c->tp == NULL) {
-		log_warn("%s", __func__);
+	buf = calloc(1, FCGI_RECORD_SIZE);
+	if (buf == NULL) {
+		log_warn("%s: calloc", __func__);
 		close(s);
 		cgi_inflight--;
 		free(c);
 		return;
 	}
 
+	c->buf = buf;
 	c->fd = s;
+	c->resp_fd = -1;
 	c->sock = sock;
-	memcpy(c->priv_fd, sock->priv_fd, sizeof(c->priv_fd));
+	memcpy(c->priv_fd, gotwebd_env->priv_fd, sizeof(c->priv_fd));
+	c->sock_id = sock->conf.id;
 	c->buf_pos = 0;
 	c->buf_len = 0;
 	c->request_started = 0;
-	c->sock->client_status = CLIENT_CONNECT;
+	c->client_status = CLIENT_CONNECT;
+	c->request_id = get_request_id();
 
-	event_set(&c->ev, s, EV_READ|EV_PERSIST, fcgi_request, c);
+	event_set(&c->ev, s, EV_READ, fcgi_request, c);
 	event_add(&c->ev, NULL);
 
 	evtimer_set(&c->tmo, fcgi_timeout, c);
 	evtimer_add(&c->tmo, &timeout);
 
-	client_cnt++;
-
+	add_request(c);
 	return;
 err:
 	cgi_inflight--;
@@ -690,14 +795,14 @@ err:
 		free(c);
 }
 
-static void
+void
 sockets_rlimit(int maxfd)
 {
 	struct rlimit rl;
 
 	if (getrlimit(RLIMIT_NOFILE, &rl) == -1)
 		fatal("%s: failed to get resource limit", __func__);
-	log_debug("%s: max open files %llu", __func__,
+	log_info("%s: max open files %llu", __func__,
 	    (unsigned long long)rl.rlim_max);
 
 	/*
