@@ -25,6 +25,7 @@
 #include <time.h>
 #include <zlib.h>
 
+#include "got_compat.h"
 #include "got_error.h"
 #include "got_repository.h"
 #include "got_object.h"
@@ -317,9 +318,11 @@ got_delta_apply_in_mem(uint8_t *base_buf, size_t base_bufsz,
 		}
 	}
 
-	if (*outsize != result_size)
-		err = got_error_msg(GOT_ERR_BAD_DELTA,
-		    "delta application result size mismatch");
+	if (err == NULL && *outsize != result_size)
+		err = got_error_fmt(GOT_ERR_BAD_DELTA,
+		    "delta application result size mismatch: actual: %zd "
+		    "expected: %llu", *outsize, result_size);
+
 	return err;
 }
 
@@ -384,12 +387,13 @@ got_delta_apply(FILE *base_file, const uint8_t *delta_buf,
 		}
 	}
 
-	if (*outsize != result_size)
-		err = got_error_msg(GOT_ERR_BAD_DELTA,
-		    "delta application result size mismatch");
+	if (err == NULL && *outsize != result_size)
+		err = got_error_fmt(GOT_ERR_BAD_DELTA,
+		    "delta application result size mismatch: actual: %zd "
+		    "expected: %llu", *outsize, result_size);
 
 	if (memstream != NULL) {
-		if (fclose(memstream) == EOF)
+		if (fclose(memstream) == EOF && err == NULL)
 			err = got_error_from_errno("fclose");
 		if (err == NULL) {
 			size_t n;
