@@ -1051,8 +1051,8 @@ recv_object_ids(struct got_object_idset *idset, struct imsgbuf *ibuf)
 }
 
 static const struct got_error *
-recv_object_id_queue(struct got_object_id_queue *queue,
-    struct got_object_idset *queued_ids, struct imsgbuf *ibuf)
+recv_object_id_queue(size_t *nids_total, struct got_object_id_queue *queue,
+    void *data, struct got_object_idset *queued_ids, struct imsgbuf *ibuf)
 {
 	const struct got_error *err = NULL;
 	int done = 0;
@@ -1060,19 +1060,26 @@ recv_object_id_queue(struct got_object_id_queue *queue,
 	struct got_object_id *ids;
 	size_t nids, i;
 
+	*nids_total = 0;
 	for (;;) {
 		err = got_privsep_recv_object_idlist(&done, &ids, &nids, ibuf);
 		if (err || done)
 			break;
+		*nids_total += nids;
 		for (i = 0; i < nids; i++) {
 			err = got_object_qid_alloc_partial(&qid);
 			if (err)
 				goto done;
 			memcpy(&qid->id, &ids[i], sizeof(qid->id));
+			if (data)
+				qid->data = data;
 			STAILQ_INSERT_TAIL(queue, qid, entry);
-			err = got_object_idset_add(queued_ids, &qid->id, NULL);
-			if (err)
-				goto done;
+			if (queued_ids) {
+				err = got_object_idset_add(queued_ids,
+				    &qid->id, NULL);
+				if (err)
+					goto done;
+			}
 		}
 		free(ids);
 		ids = NULL;
@@ -1453,7 +1460,7 @@ enumeration_request(struct imsg *imsg, struct imsgbuf *ibuf,
 	struct got_object_idset *idset, *queued_ids = NULL;
 	int i, idx, have_all_entries = 1;
 	struct enumerated_tree *trees = NULL;
-	size_t ntrees = 0, nalloc = 16;
+	size_t ntrees = 0, nalloc = 16, nids = 0;
 
 	STAILQ_INIT(&commit_ids);
 
@@ -1473,7 +1480,7 @@ enumeration_request(struct imsg *imsg, struct imsgbuf *ibuf,
 		goto done;
 	}
 
-	err = recv_object_id_queue(&commit_ids, queued_ids, ibuf);
+	err = recv_object_id_queue(&nids, &commit_ids, NULL, queued_ids, ibuf);
 	if (err)
 		goto done;
 
@@ -1670,6 +1677,136 @@ queue_commit_id(struct got_object_id_queue *ids, struct got_object_id *id,
 	return paint_commit(qid, color);
 }
 
+static const struct got_error *
+repaint_parent_commits(struct got_object_id *commit_id, int commit_idx,
+    int color, struct got_object_idset *set, struct got_object_idset *skip,
+    struct got_object_id_queue *ids, int *nids,
+    struct got_object_id_queue *painted, int *npainted,
+    struct got_pack *pack, struct got_packidx *packidx,
+    struct got_object_cache *objcache)
+{
+	const struct got_error *err = NULL;
+	const struct got_object_id_queue *parents;
+	struct got_commit_object *commit = NULL;
+	struct got_object_id_queue repaint;
+	struct got_object_idset *traversed;
+
+	STAILQ_INIT(&repaint);
+
+	traversed = got_object_idset_alloc();
+	if (traversed == NULL)
+		return got_error_from_errno("got_object_idset_alloc");
+
+	err = open_commit(&commit, pack, packidx, commit_idx, commit_id,
+	    objcache);
+	if (err)
+		return err;
+
+	while (commit) {
+		struct got_object_qid *pid, *qid;
+		int idx;
+
+		parents = got_object_commit_get_parent_ids(commit);
+		if (parents) {
+			STAILQ_FOREACH(pid, parents, entry) {
+				idx = got_packidx_get_object_idx(packidx,
+				    &pid->id);
+				/*
+				 * No need to traverse parents which are not in
+				 * the pack file, are already in the desired
+				 * set, or are marked for skipping already.
+				 */
+				if (idx == -1)
+					continue;
+				if (got_object_idset_contains(set, &pid->id))
+					continue;
+				if (set != skip &&
+				    got_object_idset_contains(skip, &pid->id))
+					continue;
+
+				if (got_object_idset_contains(traversed,
+				    &pid->id))
+					continue;
+				err = queue_commit_id(&repaint, &pid->id,
+				    color);
+				if (err)
+					goto done;
+				err = got_object_idset_add(traversed,
+				    &pid->id, NULL);
+				if (err)
+					goto done;
+			}
+		}
+		got_object_commit_close(commit);
+		commit = NULL;
+
+		pid = STAILQ_FIRST(&repaint);
+		if (pid == NULL)
+			break;
+
+		err = paint_commit(pid, color);
+		if (err)
+			break;
+
+		err = got_object_idset_add(set, &pid->id, NULL);
+		if (err)
+			break;
+
+		STAILQ_REMOVE_HEAD(&repaint, entry);
+
+		/* Insert or replace this commit on the painted list. */
+		STAILQ_FOREACH(qid, painted, entry) {
+			if (got_object_id_cmp(&qid->id, &pid->id) != 0)
+				continue;
+			err = paint_commit(qid, color);
+			if (err)
+				goto done;
+			got_object_qid_free(pid);
+			pid = qid;
+			break;
+		}
+		if (qid == NULL) {
+			STAILQ_INSERT_TAIL(painted, pid, entry);
+			(*npainted)++;
+		}
+
+		/*
+		 * In case this commit is on the caller's list of
+		 * pending commits to traverse, repaint it there.
+		 */
+		STAILQ_FOREACH(qid, ids, entry) {
+			if (got_object_id_cmp(&qid->id, &pid->id) != 0)
+				continue;
+			err = paint_commit(qid, color);
+			if (err)
+				goto done;
+			break;
+		}
+
+		idx = got_packidx_get_object_idx(packidx, &pid->id);
+		if (idx == -1) {
+			/*
+			 * Should not happen because we only queue
+			 * parents which exist in our pack file.
+			 */
+			err = got_error(GOT_ERR_NO_OBJ);
+			break;
+		}
+
+		err = open_commit(&commit, pack, packidx, idx, &pid->id,
+		    objcache);
+		if (err)
+			break;
+	}
+done:
+	if (commit)
+		got_object_commit_close(commit);
+	got_object_id_queue_free(&repaint);
+	got_object_idset_free(traversed);
+
+	return err;
+}
+
 static const struct got_error *
 paint_commits(struct got_object_id_queue *ids, int *nids,
     struct got_object_idset *keep, struct got_object_idset *drop,
@@ -1713,45 +1850,73 @@ paint_commits(struct got_object_id_queue *ids, int *nids,
 			qid = NULL;
 			continue;
 		}
+		if (color == COLOR_KEEP &&
+		    got_object_idset_contains(keep, &qid->id)) {
+			got_object_qid_free(qid);
+			qid = NULL;
+			continue;
+		}
+		if (color == COLOR_DROP &&
+		    got_object_idset_contains(drop, &qid->id)) {
+			got_object_qid_free(qid);
+			qid = NULL;
+			continue;
+		}
 
 		switch (color) {
 		case COLOR_KEEP:
-			if (got_object_idset_contains(keep, &qid->id)) {
-				got_object_qid_free(qid);
-				qid = NULL;
-				continue;
-			}
 			if (got_object_idset_contains(drop, &qid->id)) {
 				err = paint_commit(qid, COLOR_SKIP);
 				if (err)
 					goto done;
-			}
-			err = got_object_idset_add(keep, &qid->id, NULL);
-			if (err)
-				goto done;
-			break;
-		case COLOR_DROP:
-			if (got_object_idset_contains(drop, &qid->id)) {
-				got_object_qid_free(qid);
-				qid = NULL;
-				continue;
-			}
-			if (got_object_idset_contains(keep, &qid->id)) {
-				err = paint_commit(qid, COLOR_SKIP);
-				if (err)
-					goto done;
-			}
-			err = got_object_idset_add(drop, &qid->id, NULL);
-			if (err)
-				goto done;
-			break;
-		case COLOR_SKIP:
-			if (!got_object_idset_contains(skip, &qid->id)) {
 				err = got_object_idset_add(skip, &qid->id,
 				    NULL);
 				if (err)
 					goto done;
+				err = repaint_parent_commits(&qid->id, idx,
+				    COLOR_SKIP, skip, skip, ids, nids,
+				    &painted, &npainted, pack, packidx,
+				    objcache);
+				if (err)
+					goto done;
+				break;
 			}
+			if (!got_object_idset_contains(keep, &qid->id)) {
+				err = got_object_idset_add(keep, &qid->id,
+				    NULL);
+				if (err)
+					goto done;
+			}
+			break;
+		case COLOR_DROP:
+			if (got_object_idset_contains(keep, &qid->id)) {
+				err = paint_commit(qid, COLOR_SKIP);
+				if (err)
+					goto done;
+				err = got_object_idset_add(skip, &qid->id,
+				    NULL);
+				if (err)
+					goto done;
+				err = repaint_parent_commits(&qid->id, idx,
+				    COLOR_SKIP, skip, skip, ids, nids,
+				    &painted, &npainted, pack, packidx,
+				    objcache);
+				if (err)
+					goto done;
+				break;
+			}
+			if (!got_object_idset_contains(drop, &qid->id)) {
+				err = got_object_idset_add(drop, &qid->id,
+				    NULL);
+				if (err)
+					goto done;
+			}
+			break;
+		case COLOR_SKIP:
+			err = got_object_idset_add(skip, &qid->id,
+			    NULL);
+			if (err)
+				goto done;
 			break;
 		default:
 			/* should not happen */
@@ -1870,24 +2035,47 @@ commit_painting_request(struct imsg *imsg, struct imsgbuf *ibuf,
     struct got_object_idset *drop, struct got_object_idset *skip)
 {
 	const struct got_error *err = NULL;
-	struct got_imsg_commit_painting_request ireq;
-	struct got_object_id id;
 	size_t datalen;
 	struct got_object_id_queue ids;
+	size_t nkeep = 0, ndrop = 0, nskip = 0;
 	int nids = 0;
+	uintptr_t color;
 
 	STAILQ_INIT(&ids);
 
 	datalen = imsg->hdr.len - IMSG_HEADER_SIZE;
-	if (datalen != sizeof(ireq))
+	if (datalen != 0)
 		return got_error(GOT_ERR_PRIVSEP_LEN);
-	memcpy(&ireq, imsg->data, sizeof(ireq));
-	memcpy(&id, &ireq.id, sizeof(id));
 
-	err = queue_commit_id(&ids, &id, ireq.color);
+	color = COLOR_KEEP;
+	err = recv_object_id_queue(&nkeep, &ids, (void *)color, NULL, ibuf);
 	if (err)
-		return err;
-	nids = 1;
+		goto done;
+	if (nids + nkeep > INT_MAX) {
+		err = got_error(GOT_ERR_PRIVSEP_LEN);
+		goto done;
+	}
+	nids += nkeep;
+
+	color = COLOR_DROP;
+	err = recv_object_id_queue(&ndrop, &ids, (void *)color, NULL, ibuf);
+	if (err)
+		goto done;
+	if (nids + ndrop > INT_MAX) {
+		err = got_error(GOT_ERR_PRIVSEP_LEN);
+		goto done;
+	}
+	nids += ndrop;
+
+	color = COLOR_SKIP;
+	err = recv_object_id_queue(&nskip, &ids, (void *)color, NULL, ibuf);
+	if (err)
+		goto done;
+	if (nids + nskip > INT_MAX) {
+		err = got_error(GOT_ERR_PRIVSEP_LEN);
+		goto done;
+	}
+	nids += nskip;
 
 	err = paint_commits(&ids, &nids, keep, drop, skip,
 	    pack, packidx, ibuf, objcache);
